from pathlib import Path

from fastapi import HTTPException
from fastapi import status as s
from loguru import logger
from redbaby.pyobjectid import PyObjectId

from ..authz.engines.factory import AuthorizationEngine
from ..authz.policies.controllers import upsert_one
from ..authz.policies.schemas import AuthorizationPolicyIn
from ..schemas import Infostar

DEFAULT_POLICIES = {
    "melt-key": {
        "path": Path(__file__).parents[2] / "resources" / "policies" / "melt-key.rego",
        "description": "MELT API Key privilege levels.",
    },
}

SYSTEM_INFOSTAR = Infostar(
    request_id=PyObjectId(),
    apikey_name="default",
    authprovider_org="/",
    authprovider_type="melt-key",
    extra={},
    service_handle="tauth",
    user_handle="sysadmin@teialabs.com",
    user_owner_handle="/",
    original=None,
    client_ip="127.0.0.1",
)


def init_app():
    logger.debug("Initializing policy engine.")
    authz_engine = AuthorizationEngine.get()  # caching
    # TODO: maybe move this to authz engine
    for name, policy_data in DEFAULT_POLICIES.items():
        logger.debug(f"Loading policy: {name!r} from {policy_data['path']!r}.")
        policy_content = policy_data["path"].read_text()
        policy = AuthorizationPolicyIn(
            name=name,
            description=policy_data["description"],
            policy=policy_content,
            type="opa",
        )
        try:
            upsert_one(policy, SYSTEM_INFOSTAR)
        except HTTPException as e:
            if e.status_code != s.HTTP_409_CONFLICT:
                raise e
            logger.debug(f"Policy {name} already exists. Skipping.")

    logger.debug("Policy engine initialized.")



from datetime import datetime


def validate_id(key, obj: dict):
    assert key in obj
    assert isinstance(obj[key], str)
    # TODO: assert encoding and hash


def validate_float(key, obj: dict):
    assert key in obj
    assert isinstance(obj[key], float)


def validate_int(key, obj: dict):
    assert key in obj
    assert isinstance(obj[key], int)


def validate_nonzero_int(key: str, obj: dict):
    validate_int(key, obj)
    assert key in obj
    assert isinstance(obj[key], int)
    assert obj[key] > 0


def validate_nonzero(key: str, obj: dict):
    assert key in obj
    assert obj[key] > 0


def validate_nonzero_float(key: str, obj: dict):
    assert key in obj
    assert isinstance(obj[key], float)
    assert obj[key] > 0


def validate_isostring(key: str, obj: dict):
    assert key in obj
    assert isinstance(obj[key], str)
    dt = datetime.fromisoformat(obj[key])
    assert isinstance(dt, datetime)


def run_validations(obj: dict, validations: dict) -> None:
    print(obj)
    print(type(validations))
    print(validations)
    print()
    for key, validation in validations.items():
        if isinstance(obj[key], list) and callable(validation):
            for item in obj[key]:
                run_validations(item, validation)
            continue
        elif isinstance(obj[key], list) and isinstance(validation, list):
            assert len(obj[key]) == len(validation)
            for item, val_item in zip(obj[key], validation):
                run_validations(item, val_item)
            continue
        if isinstance(obj[key], dict):
            run_validations(obj[key], validation)
            continue
        if callable(validation):
            validation(key, obj)
        else:
            print(key, obj[key], validation, sep="\n\n")
            assert obj[key] == validation


def validate_token(key, obj):
    value = obj[key]
    assert isinstance(value, str)
    assert value.startswith("MELT_")
    pieces = value.split("--")
    assert len(pieces) == 3
    assert len(pieces[-1]) >= 16


def validate_nonempty_string(key, obj):
    assert key in obj
    assert isinstance(obj[key], str)
    assert len(obj[key]) > 0


def validate_chatml_role(key, obj):
    assert key in obj
    assert isinstance(obj[key], str)
    assert obj[key] in {"system", "user", "assistant"}



"""
FastAPI dependency injection for privilege checking.

is_valid_user - anyone with a valid key
is_valid_admin - token_name == default
is_valid_superuser - token_name == default and client_name == $root
"""

from typing import Callable

from fastapi import HTTPException, Request, status

from ..schemas import Creator


def get_creator(
    validate_access: Callable[[Creator], bool]
) -> Callable[[Request], Creator]:
    def wrapper(request: Request) -> Creator:
        c = request.state.creator
        if validate_access(c):
            return c
        else:
            s = status.HTTP_403_FORBIDDEN
            d = {  # TODO: use http_error_schemas
                "msg": "You do not have access to this resource.",
                "info": {"creator": c.dict()},
            }
            # TODO: delegate exception raising to the wrapped function
            raise HTTPException(status_code=s, detail=d)

    return wrapper


@get_creator
def is_valid_user(creator: Creator) -> bool:
    return True


@get_creator
def is_valid_admin(creator: Creator) -> bool:
    return creator.token_name == "default"


@get_creator
def is_valid_superuser(creator: Creator) -> bool:
    return creator.token_name == "default" and creator.client_name == "/"


@get_creator
def is_direct_user(creator: Creator) -> bool:
    return creator.token_name not in ("auth0-jwt", "azure-jwt")



"""Generate API reference pages and navigation."""

from pathlib import Path

import mkdocs_gen_files


def build_reference_docs(library_path: Path):
    """
    Builds the reference documentation for the library.

    Automatically generates the reference documentation for Python modules
    and packages and creates a navigation file.

    Args:
        library_path: path to the library's root directory.
    """
    nav = mkdocs_gen_files.Nav()

    for path in sorted(library_path.rglob("*.py")):
        module_path = path.relative_to(library_path.parent).with_suffix("")
        doc_path = path.relative_to(library_path.parent).with_suffix(".md")
        full_doc_path = Path("reference", doc_path)

        parts = tuple(module_path.parts)

        if parts[-1] == "__init__":
            parts = parts[:-1]
            doc_path = doc_path.with_name("index.md")
            full_doc_path = full_doc_path.with_name("index.md")
        elif parts[-1] == "__main__":
            continue

        nav[parts] = doc_path.as_posix()

        with mkdocs_gen_files.open(full_doc_path, "w") as fd:
            ident = ".".join(parts)
            fd.write(f"::: {ident}")

        mkdocs_gen_files.set_edit_path(full_doc_path, path)

    with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
        nav_file.writelines(nav.build_literate_nav())


# Change this to point to the package's root directory
library_path = Path(__file__).parents[2] / "tauth"
build_reference_docs(library_path=library_path)



import subprocess
import sys

import uvicorn

from .settings import Settings


def main():
    settings = Settings()

    # Caching API
    subprocess.Popen(
        "python -m cacheia_api",
        shell=True,
        stdout=sys.stdout,
        stderr=sys.stderr,
    )

    # Authorization API
    # improved debug logging: --log-format=json-pretty
    subprocess.Popen(
        "./opa run --server --log-level=debug --v1-compatible",
        shell=True,
        stdout=sys.stdout,
        stderr=sys.stderr,
    )

    uvicorn.run(
        app="tauth.app:create_app",
        factory=True,
        forwarded_allow_ips="*",
        host=settings.HOST,
        port=settings.PORT,
        reload=settings.RELOAD,
        workers=settings.WORKERS,
    )


if __name__ == "__main__":
    main()



import re
import secrets
from functools import lru_cache

from fastapi import HTTPException
from fastapi import status as s
from http_error_schemas.schemas import RequestValidationError
from multiformats import multibase

from ...settings import Settings
from ..melt_key.models import TokenDAO


def parse_token(token_value: str) -> tuple[str, str, str]:
    """
    Parse token string into client name, token name, and token value.

    Raise an error if token is incorrectly formatted.
    >>> parse_token("MELT_/client-name--token-name--abcdef123456789")
    ('client-name', 'token-name', 'abcdef123456789')
    """
    stripped = token_value.lstrip("MELT_")
    pieces = stripped.split("--")
    if len(pieces) != 3:
        code, m = 401, "Token is not in the correct format."
        raise HTTPException(status_code=code, detail=m)
    return tuple(pieces)  # type: ignore


def create_token(client_name: str, token_name: str):
    token_value = multibase.encode(secrets.token_bytes(24), "base58btc")
    fmt_token_value = f"MELT_{client_name}--{token_name}--{token_value}"
    return fmt_token_value


@lru_cache(maxsize=32)
def validate_token_against_db(token: str, client_name: str, token_name: str):
    filters = {"client_name": client_name, "name": token_name}
    entity = TokenDAO.collection(alias=Settings.get().TAUTH_REDBABY_ALIAS).find_one(
        filter=filters
    )
    if not entity:
        d = {
            "filters": filters,
            "msg": f"Token does not exist for client.",
            "type": "DocumentNotFound",
        }
        raise HTTPException(status_code=s.HTTP_401_UNAUTHORIZED, detail=d)
    if not secrets.compare_digest(token, entity["value"]):
        code, m = s.HTTP_401_UNAUTHORIZED, "Token does not match."
        raise HTTPException(status_code=code, detail={"msg": m})
    return entity


def sanitize_client_name(client_name: str, loc: list[str] = ["body", "name"]) -> str:
    print(client_name)
    if client_name != "/":
        clean_client_name = client_name.rstrip("/").lower()
    else:
        clean_client_name = client_name
    if not clean_client_name.startswith("/"):
        details = RequestValidationError(
            loc=loc,
            msg="Client names are namespaced and must be absolute (start with a slash character).",
            type="InvalidClientName",
        )
        raise HTTPException(status_code=s.HTTP_422_UNPROCESSABLE_ENTITY, detail=details)
    if pos := re.search(r"\s", clean_client_name):
        details = RequestValidationError(
            loc=loc + [str(pos)],
            msg="Client name cannot contain spaces.",
            type="InvalidClientName",
        )
        raise HTTPException(status_code=s.HTTP_422_UNPROCESSABLE_ENTITY, detail=details)
    if pos := re.search(r"//", clean_client_name):
        details = RequestValidationError(
            loc=loc + [str(pos)],
            msg="Client name cannot contain consecutive slashes.",
            type="InvalidClientName",
        )
        raise HTTPException(status_code=s.HTTP_422_UNPROCESSABLE_ENTITY, detail=details)
    if pos := re.search(r"--", clean_client_name):
        details = RequestValidationError(
            loc=loc + [str(pos)],
            msg="Client name cannot contain consecutive dashes. Single dashes are fine.",
            type="InvalidClientName",
        )
        raise HTTPException(status_code=s.HTTP_422_UNPROCESSABLE_ENTITY, detail=details)
    return clean_client_name



from fastapi import FastAPI

from ..settings import Settings
from . import authorization, cache, database, security


def init_app(app: FastAPI, sets: Settings) -> None:
    cache.init_app()
    database.init_app(sets)
    security.init_app(app)
    authorization.init_app()



from ...settings import Settings
from .interface import AuthorizationInterface
from .opa import OPAEngine


class AuthorizationEngine:
    _instance: AuthorizationInterface | None = None

    @classmethod
    def get(cls) -> AuthorizationInterface:
        if not cls._instance:
            # settings = Settings.get()
            cls._instance = OPAEngine()
        return cls._instance



from fastapi import HTTPException
from fastapi import status as s
from loguru import logger

from ...schemas import Infostar
from ...schemas.gen_fields import GeneratedFields
from ...settings import Settings
from ...utils import creation, reading
from ..engines.factory import AuthorizationEngine
from ..policies.models import AuthorizationPolicyDAO
from ..policies.schemas import AuthorizationPolicyIn


def upsert_one(
    body: AuthorizationPolicyIn,
    infostar: Infostar,
) -> GeneratedFields:
    # Insert policy in TAuth's database
    logger.debug("Inserting policy in TAuth.")
    policy_col = AuthorizationPolicyDAO.collection(
        alias=Settings.get().TAUTH_REDBABY_ALIAS
    )
    policy_content = body.model_dump(by_alias=True) | {
        "created_by": infostar.model_dump()
    }
    result = policy_col.update_one(
        {"name": body.name},
        {"$set": policy_content},
        upsert=True,
    )
    logger.debug(f"Policy upsert result: {result}.")
    policy = AuthorizationPolicyDAO(**policy_content)

    # Insert policy in authorization provider
    logger.debug("Inserting policy in AuthZ provider.")
    authz_engine = AuthorizationEngine.get()
    try:
        result = authz_engine.upsert_policy(
            policy_name=body.name,
            policy_content=body.policy,
        )
    except Exception as e:  # TODO: exception abstraction for authz provider errors
        raise HTTPException(
            status_code=s.HTTP_400_BAD_REQUEST,
            detail=dict(msg=f"Failed to create policy {body.name!r} in engine: {e}"),
        )
    if not result:
        logger.debug(f"Failed to create policy in authorization engine.")
        authz_policy_col = AuthorizationPolicyDAO.collection()
        result = authz_policy_col.delete_one({"name": body.name})
        logger.debug(f"Deleted objects from TAuth DB: {result.deleted_count}.")
        raise HTTPException(
            status_code=s.HTTP_400_BAD_REQUEST,
            detail=dict(msg=f"Failed to create policy {body.name!r} in engine."),
        )

    logger.debug(f"Inserted policy in authorization engine.")
    return GeneratedFields(**policy.model_dump(by_alias=True))


def read_one(id: str) -> AuthorizationPolicyDAO:
    logger.debug("Reading policy from TAuth.")
    policy = reading.read_one(
        infostar={},  # type: ignore
        model=AuthorizationPolicyDAO,
        identifier=id,
    )
    print(type(policy))
    return policy


def read_many(
    filters: dict,
    infostar: Infostar,
) -> list[AuthorizationPolicyDAO]:
    logger.debug("Reading policies from TAuth.")
    policy = reading.read_many(
        infostar=infostar,
        model=AuthorizationPolicyDAO,
        **filters,
    )
    return policy


def delete_one(id: str) -> None:
    logger.debug("Deleting policy from TAuth.")
    policy = reading.read_one(
        infostar={},  # type: ignore
        model=AuthorizationPolicyDAO,
        identifier=id,
    )
    policy_col = AuthorizationPolicyDAO.collection()
    result = policy_col.delete_one({"name": policy.name})
    logger.debug(f"Deleted objects from TAuth DB: {result.deleted_count}.")

    logger.debug("Deleting policy from AuthZ provider.")
    authz_engine = AuthorizationEngine.get()
    try:
        result = authz_engine.delete_policy(
            policy_name=policy.name,
        )
    except Exception as e:  # TODO: exception abstraction for authz provider errors
        raise HTTPException(
            status_code=s.HTTP_400_BAD_REQUEST,
            detail=dict(msg=f"Failed to delete policy {policy.name!r} in engine: {e}"),
        )
    if not result:
        logger.debug(f"Failed to delete policy in authorization engine.")
        raise HTTPException(
            status_code=s.HTTP_400_BAD_REQUEST,
            detail=dict(msg=f"Failed to delete policy {policy.name!r} in engine."),
        )
    logger.debug(f"Deleted policy from authorization engine.")



from importlib.metadata import version

from fastapi import APIRouter, FastAPI

from . import dependencies
from .authn.routes import router as authentication_router
from .authproviders import router as authproviders_router
from .authz.policies.routes import router as policy_router
from .authz.routes import router as authorization_router
from .entities import router as entities_router
from .legacy import client, tokens
from .settings import Settings


def create_app() -> FastAPI:
    settings = Settings()
    app = FastAPI(
        title="TAuth",
        description="**T**eia **Auth**entication Service.",
        version=version("tauth"),
    )
    dependencies.init_app(app, settings)

    # Routes
    @app.get("/", status_code=200, tags=["health 🩺"])
    def _():
        return {"status": "ok"}

    router = APIRouter()
    router.include_router(get_router())
    app.include_router(router)

    return app


def get_router() -> APIRouter:
    base_router = APIRouter()
    base_router.include_router(authentication_router)
    base_router.include_router(authorization_router)
    base_router.include_router(policy_router)
    base_router.include_router(entities_router)
    base_router.include_router(authproviders_router)
    base_router.include_router(client.router)
    base_router.include_router(tokens.router)
    return base_router



import pytest
from fastapi.testclient import TestClient
from pymongo.database import Database

from .utils import run_validations, validate_id, validate_isostring


@pytest.fixture(autouse=True)
def teardown(tauth_db: Database):
    try:
        yield None
    finally:
        tauth_db["organizations"].delete_many({"name": "/teia"})


@pytest.fixture(scope="module")
def organization_obj() -> dict:
    return {
        "name": "/teia",
        "external_ids": [
            {
                "name": "oauth_org_id",
                "value": "org_123",
            },
        ]
    }


def test_create_one(
    client: TestClient,
    headers: dict,
    organization_obj: dict,
) -> None:
    response = client.post("/api/tauth/organizations", json=organization_obj, headers=headers)
    assert response.status_code == 201
    print(response.json())
    response_body = response.json()
    response_body["created_at"] = validate_isostring("created_at", response_body)


from pathlib import Path

from fastapi import APIRouter, Depends, Request
from fastapi import status as s

from ..dependencies.security import RequestAuthenticator
from ..schemas import Infostar

service_name = Path(__file__).parent.name
router = APIRouter(prefix=f"/{service_name}", tags=[service_name + " 🪪"])


@router.post("", status_code=s.HTTP_200_OK)
@router.post("/", status_code=s.HTTP_200_OK, include_in_schema=False)
async def authenticate(request: Request) -> Infostar:
    infostar: Infostar = request.state.infostar
    return infostar



from typing import Any

from jwt import PyJWKSet
from pytest_mock import MockerFixture

from tauth.auth.auth0_dyn import RequestAuthenticator


class AuthProviderMock:
    def get_external_id(self, *_) -> str:
        return "https://test/"


class StateMock:
    def __init__(self) -> None:
        self.keys = {}

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "keys":
            object.__setattr__(self, name, value)
        else:
            self.keys[name] = value

    def __getattribute__(self, name: str) -> Any:
        if name == "keys":
            return object.__getattribute__(self, name)
        return self.keys.get(name)


class RequestMock:
    def __init__(self) -> None:
        self.state = StateMock()
        self.headers = {}
        self.client = None


def test_auth0_dyn(access_token: str, id_token: str, jwk: dict, mocker: MockerFixture):
    provider_target_fn = "tauth.auth.auth0_dyn.RequestAuthenticator.get_authprovider"
    mocker.patch(target=provider_target_fn, new=lambda *_, **__: AuthProviderMock)

    jwk_target_fn = "tauth.auth.auth0_dyn.ManyJSONKeySetStore.get_jwks"
    mocker.patch(target=jwk_target_fn, new=lambda *_, **__: PyJWKSet.from_dict(jwk))

    org_db_query_fn = "tauth.organizations.controllers.read_one"
    mocker.patch(target=org_db_query_fn, new=lambda *_, **__: {"name": "test"})

    request = RequestMock()

    RequestAuthenticator.validate(request, access_token, id_token)  # type: ignore
    assert True



from tauth.utils.fastapi_extension import setup_database

from ..settings import Settings


def init_app(sets: Settings):
    setup_database(
        dbname=sets.TAUTH_MONGODB_DBNAME,
        dburi=sets.TAUTH_MONGODB_URI,
        redbaby_alias=sets.TAUTH_REDBABY_ALIAS,
    )



from cacheia_api.settings import Settings as CacheiaSettings
from cacheia_client import configure


def init_app():
    sets = CacheiaSettings()
    configure(url=f"http://{sets.CACHEIA_HOST}:{sets.CACHEIA_PORT}")



from pathlib import Path
from typing import TypedDict

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from jwcrypto import jwk

from .constants import DEFAULT_PATH


class JWK(TypedDict, total=False):
    alg: str
    use: str
    kty: str
    n: str
    e: str
    kid: str


class JWKS(TypedDict):
    keys: list[JWK]


def generate_jwks(public_key: rsa.RSAPublicKey) -> JWKS:
    # Convert the public key PEM to a JWK object
    jwk_public_key = jwk.JWK.from_pem(public_key_bytes(public_key))
    dict_jwk: dict[str, str] = jwk_public_key.export(as_dict=True)  # type: ignore

    public_jwk = JWKS(keys=[JWK(alg="RS256", use="sig", **dict_jwk)])
    return public_jwk


def generate_signing_pair(
    key_size: int = 2048,
) -> tuple[rsa.RSAPrivateKey, rsa.RSAPublicKey]:
    """
    Generate an RSA private key and extract the public key.
    """

    # Generate an RSA private key
    private_key = rsa.generate_private_key(
        key_size=key_size,
        public_exponent=65537,
        backend=default_backend(),
    )

    # Extract the public key from the private key
    public_key = private_key.public_key()

    return private_key, public_key


def private_key_bytes(private_key: rsa.RSAPrivateKey) -> bytes:
    """
    Serialize the private key to PEM format.
    """

    # Serialize the private key to PEM format
    private_bytes = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return private_bytes


def public_key_bytes(public_key: rsa.RSAPublicKey) -> bytes:
    """
    Serialize the public key to PEM format.
    """

    # Serialize the public key to PEM format
    public_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return public_bytes


def save_signing_pair(
    private_key: rsa.RSAPrivateKey,
    public_key: rsa.RSAPublicKey,
    path: Path = DEFAULT_PATH,
) -> tuple[Path, Path]:
    """
    Save the private and public keys to disk.
    """

    if not path.exists():
        path.mkdir(parents=True)

    private_path = path / "private_key.pem"
    public_path = path / "public_key.pem"

    private_bytes = private_key_bytes(private_key)

    # Write the private key PEM file to disk
    with open(private_path, "wb") as pem_file:
        pem_file.write(private_bytes)

    public_bytes = public_key_bytes(public_key)

    # Write the public key PEM file to disk
    with open(public_path, "wb") as pem_file:
        pem_file.write(public_bytes)

    return private_path, public_path



from typing import Literal

from pymongo import IndexModel
from redbaby.behaviors.objectids import ObjectIdMixin
from redbaby.behaviors.reading import ReadingMixin
from redbaby.document import Document

from ...utils.teia_behaviors import Authoring


class AuthorizationPolicyDAO(Document, Authoring, ObjectIdMixin, ReadingMixin):
    description: str
    name: str
    policy: str
    type: Literal["opa"]

    @classmethod
    def collection_name(cls) -> str:
        return "authorization-policies"

    @classmethod
    def indexes(cls) -> list[IndexModel]:
        idxs = [
            IndexModel([("name", 1), ("type", 1)], unique=True),
            IndexModel([("type", 1)]),
        ]
        return idxs

    def hashable_fields(self) -> list[str]:
        return [self.name, self.type]



from typing import Literal, Optional, Self

from pydantic import BaseModel, Field, model_validator
from pydantic.config import ConfigDict
from redbaby.pyobjectid import PyObjectId

from ..schemas.attribute import Attribute
from .models import OrganizationRef, ServiceRef


class AuthProviderIn(BaseModel):
    external_ids: list[Attribute] = Field(default_factory=list)
    extra: list[Attribute] = Field(default_factory=list)
    organization_name: str
    service_name: Optional[str] = Field(None)
    type: Literal["auth0", "melt-key", "tauth-key"]

    @model_validator(mode="after")
    def check_external_ids(self: Self) -> Self:
        if self.type == "auth0":
            for field in self.external_ids:
                if field.name == "issuer":
                    if not field.value.startswith("https://"):
                        raise ValueError("Issuer must be an HTTPS URL.")
                    if not field.value.endswith("/"):
                        field.value += "/"
        return self

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "melt-api-key": {
                        "summary": "MELT API Key (Teia Labs)",
                        "description": "MELT API Keys of the form `MELT_<organization>/[service]--<key_name>--<key_id>`.",
                        "value": {
                            "organization_name": "/teialabs",
                            # "service_name": "",
                            "type": "melt-key",
                        },
                    },
                    "auth0": {
                        "summary": "Auth0 (Teia Labs)",
                        "description": "Teia Labs' Auth0 as an authprovider.",
                        "value": {
                            "external_ids": [
                                {
                                    "name": "issuer",
                                    "value": "https://dev-z60iog20x0slfn0a.us.auth0.com/",
                                },
                                {
                                    "name": "audience",
                                    "value": "api://allai.chat.webui",
                                },
                            ],
                            "organization_name": "/teialabs",
                            "service_name": "athena-api",
                            "type": "auth0",
                        },
                    },
                },
            ]
        }
    )


class AuthProviderMoreIn(BaseModel):
    external_ids: list[Attribute] = Field(default_factory=list)
    extra: list[Attribute] = Field(default_factory=list)
    organization_ref: OrganizationRef
    service_ref: Optional[ServiceRef]
    type: Literal["auth0", "melt-key", "tauth-key"]


class AuthProviderRef(BaseModel):
    id: PyObjectId = Field(alias="_id")
    organizaion_ref: OrganizationRef
    service_ref: Optional[ServiceRef]
    type: Literal["auth0", "melt-key", "tauth-key"]



from abc import ABC, abstractmethod
from typing import Optional

from ...entities.models import EntityDAO


class AuthorizationInterface(ABC):
    @abstractmethod
    def is_authorized(
        self,
        entity: EntityDAO,
        policy_name: str,
        resource: str,
        context: Optional[dict] = None,
    ) -> bool: ...

    @abstractmethod
    def get_filters(
        self,
        entity: EntityDAO,
        policy_name: str,
        resource: str,
        context: Optional[dict] = None,
    ) -> dict: ...

    @abstractmethod
    def upsert_policy(self, policy_name: str, policy_content: str) -> bool: ...

    @abstractmethod
    def delete_policy(self, policy_name: str) -> bool: ...



from datetime import datetime, timezone

from fastapi import APIRouter, Body, Depends, HTTPException, Request
from fastapi import status as s
from http_error_schemas.schemas import RequestValidationError
from loguru import logger

from ..authn.melt_key.authorization import (
    validate_creation_access_level,
    validate_scope_access_level,
)
from ..authn.melt_key.models import TokenDAO
from ..authn.melt_key.schemas import (
    ClientCreation,
    ClientCreationOut,
    ClientOut,
    ClientOutJoinTokensAndUsers,
    TokenCreationIntermediate,
    TokenCreationOut,
    TokenMeta,
    UserOut,
)
from ..authn.melt_key.token import create_token
from ..authproviders.models import AuthProviderDAO
from ..authz import privileges
from ..entities.models import EntityDAO
from ..schemas import Creator, Infostar
from ..settings import Settings
from ..utils import creation, reading

router = APIRouter(prefix="/clients", tags=["legacy"])


@router.post("", status_code=s.HTTP_201_CREATED)
@router.post("/", status_code=s.HTTP_201_CREATED, include_in_schema=False)
async def create_one(
    request: Request,
    client_in: ClientCreation = Body(),
    infostar: Infostar = Depends(privileges.is_valid_admin),
) -> ClientCreationOut:
    """
    Create a new client.

    - A client with name `/teia/athena` can only be created by `/teia` or `/`.
    - In order to create `/teia/athena/chat`, you must first create `/teia/athena`.
    - Trailing slashes are ignored.
    """
    creator: Creator = request.state.creator
    logger.debug(f"Creating a new client: {client_in.name!r}.")
    logger.debug(f"Validating access: {creator.client_name!r} -> {client_in.name!r}.")
    # TODO: move this to rego
    validate_creation_access_level(client_in.name, creator.client_name)
    try:
        org_handle = client_in.name.split("/")[1]
        organization = f"/{org_handle}"
        logger.debug(f"Checking if organization entity {organization!r} exists.")
        filters = {"handle": organization, "type": "organization"}
        org = reading.read_one_filters(infostar, model=EntityDAO, **filters)
    except HTTPException as e:
        details = RequestValidationError(
            loc=["body", "name"],
            msg="Cannot create client for non-existent organization entity.",
            type="DocumentNotFound",
        )
        raise HTTPException(status_code=s.HTTP_404_NOT_FOUND, detail=details)

    try:
        logger.debug(f"Checking if org {org.handle!r} has 'melt-key' authprovider.")
        filters = {"type": "melt-key", "organization_ref.handle": org.handle}
        provider = reading.read_one_filters(infostar, model=AuthProviderDAO, **filters)
    except HTTPException as e:
        details = RequestValidationError(
            loc=["body", "name"],
            msg="Cannot create client for organization with no 'melt-key' authprovider.",
            type="DocumentNotFound",
        )
        raise HTTPException(status_code=s.HTTP_404_NOT_FOUND, detail=details)

    logger.debug(f"Checking if authprovider has necessary subclients.")
    existing_clients = set(
        sorted([e.value for e in provider.extra if e.name == "melt_key_client"])
    )
    logger.debug(f"Existing clients for authprovider: {existing_clients!r}.")
    needed_clients = [c for c in client_in.name.split("/")[:-1] if c != ""]
    needed_clients = [
        "/" + "/".join(needed_clients[: i + 1]) for i, _ in enumerate(needed_clients)
    ]
    logger.debug(f"Needed clients to create new client: {needed_clients!r}.")
    missing_clients = set(needed_clients) - existing_clients
    if missing_clients:
        details = RequestValidationError(
            loc=["body", "name"],
            msg=f"Cannot create client due to non-existent parent client(s): {missing_clients!r}.",
            type="DocumentNotFound",
        )
        raise HTTPException(status_code=s.HTTP_404_NOT_FOUND, detail=details)

    try:
        token_name = "default"
        logger.debug(f"Creating {token_name!r} token.")
        token = TokenCreationIntermediate(
            client_name=client_in.name,
            name=token_name,
            value=create_token(client_in.name, token_name),
        )
        token = creation.create_one(token, model=TokenDAO, infostar=infostar)
        token_out = TokenCreationOut(
            **token.model_dump(exclude={"created_by"}), created_by=creator
        )
    except HTTPException as e:
        details = RequestValidationError(
            loc=["body", "name"],
            msg="Client already exists.",
            type="DuplicateKeyError",
        )
        raise HTTPException(status_code=s.HTTP_409_CONFLICT, detail=details)

    logger.debug(f"Adding new subclient to authprovider.")
    authprovider_collection = AuthProviderDAO.collection(
        alias=Settings.get().TAUTH_REDBABY_ALIAS
    )
    authprovider_collection.update_one(
        filter={"_id": provider.id},
        update={
            "$addToSet": {
                "extra": {
                    "name": "melt_key_client",
                    "value": client_in.name,
                }
            }
        },
    )

    out = ClientCreationOut(
        created_at=datetime.now(timezone.utc),
        created_by=creator,
        _id=org.id,
        name=client_in.name,
        tokens=[token_out],
        users=[],
    )
    return out


@router.get("", status_code=s.HTTP_200_OK)
@router.get("/", status_code=s.HTTP_200_OK, include_in_schema=False)
async def read_many(request: Request) -> list[ClientOut]:
    creator: Creator = request.state.creator
    infostar: Infostar = request.state.infostar
    logger.debug(f"Fetching clients for client: {creator.client_name!r}.")

    org_handle = creator.client_name.split("/")[1]
    organization = f"/{org_handle}"
    logger.debug(f"Organization entity handle: {organization!r}.")

    # Root token has access to multiple authproviders
    logger.debug(f"Reading 'melt-key' authprovider(s) for {organization!r}.")
    filters = {
        "type": "melt-key",
        "organization_ref.handle": {"$regex": f"^{organization}"},
    }
    providers = reading.read_many(infostar, model=AuthProviderDAO, **filters)

    logger.debug(f"Finding all clients for authprovider(s).")
    clients = []
    for provider in providers:
        logger.debug(f"Reading clients for authprovider {provider.id!r}.")
        for e in provider.extra:
            if e.name == "melt_key_client":
                # We need to filter the clients depending on the subclient level
                if creator.client_name not in e.value:
                    continue
                clients.append(
                    ClientOut(
                        created_at=provider.created_at,
                        created_by=creator,
                        _id=str(provider.id),
                        name=e.value,
                    )
                )

    return clients


@router.get("/{name:path}", status_code=s.HTTP_200_OK)
async def read_one(request: Request, name: str) -> ClientOutJoinTokensAndUsers:
    creator: Creator = request.state.creator
    infostar: Infostar = request.state.infostar
    logger.debug(f"Reading client {name!r}.")

    logger.debug(f"Validating access: {creator.client_name!r} -> {name!r}.")
    validate_scope_access_level(name, creator.client_name)

    logger.debug(f"Reading authprovider for client {name!r}.")
    reading.read_one_filters(
        infostar=infostar,
        model=AuthProviderDAO,
        type="melt-key",
        **{
            "extra.name": "melt_key_client",
            "extra.value": name,
        },
    )
    logger.debug(f"Reading tokens for client {name!r}.")
    filters = {"client_name": name}
    tokens: list[TokenDAO] = reading.read_many(infostar, model=TokenDAO, **filters)
    tokens_view = [
        TokenMeta(
            **t.model_dump(exclude={"created_by"}),
            created_by=Creator.from_infostar(t.created_by),
        )
        for t in tokens
    ]

    org_handle = name.split("/")[1]
    organization = f"/{org_handle}"
    logger.debug(f"Reading organization entity: {organization!r}.")
    filters = {"type": "organization", "handle": organization}
    org = reading.read_one_filters(infostar, model=EntityDAO, **filters)

    logger.debug(f"Reading users for organization entity {org.handle!r}.")
    users = reading.read_many(
        infostar, EntityDAO, **{"type": "user", "owner_ref.handle": org.handle}
    )

    client_view = ClientOutJoinTokensAndUsers(
        created_at=org.created_at,
        created_by=creator,
        _id=org.id,  # TODO: Maybe this should be the authprovider ID  nei: kkkk
        name=name,
        tokens=tokens_view,
        users=[
            UserOut(
                **u.model_dump(exclude={"created_by"}),
                created_by=creator,
                email=u.handle,
            )
            for u in users
        ],
    )
    return client_view



from fastapi import Depends
from redbaby.database import DB
from redbaby.document import Document

from tauth.dependencies.security import RequestAuthenticator


def setup_database(dbname: str, dburi: str, redbaby_alias: str):
    DB.add_conn(
        db_name=dbname,
        uri=dburi,
        alias=redbaby_alias,
    )
    for m in Document.__subclasses__():
        if m.__module__.startswith("tauth"):
            m.create_indexes(alias=redbaby_alias)


def get_depends():
    return Depends(RequestAuthenticator.validate)



from typing import Literal, Optional

from pydantic import BaseModel, Field
from pydantic.config import ConfigDict

from ..schemas.attribute import Attribute
from .models import EntityRef


class EntityIn(BaseModel):
    external_ids: list[Attribute] = Field(default_factory=list)
    extra: list[Attribute] = Field(default_factory=list)
    handle: str = Field(..., min_length=3, max_length=50)
    owner_handle: Optional[str] = Field(None)
    roles: list[str] = Field(default_factory=list)
    type: Literal["user", "service", "organization"]

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {
                    "org": {
                        "summary": "Minimal Organization",
                        "description": "A root-level organization with no authproviders registered.",
                        "value": {
                            "handle": "/orgname",
                            "owner_handle": None,
                            "type": "organization",
                        },
                    },
                    "org-user": {
                        "summary": "Minimal Organization User",
                        "description": 'A user registered in an organization. "owner_handle" must point to a valid organization handle.',
                        "value": {
                            "handle": "user@orgname.com",
                            "owner_handle": "/orgname",
                            "type": "user",
                        },
                    },
                }
            ]
        }
    )


class EntityIntermediate(BaseModel):
    external_ids: list[Attribute] = Field(default_factory=list)
    extra: list[Attribute] = Field(default_factory=list)
    handle: str = Field(...)
    owner_ref: Optional[EntityRef] = Field(None)
    roles: list[str] = Field(default_factory=list)
    type: Literal["user", "service", "organization"]



from typing import Type, TypeVar

from cacheia_decorators.remote import cache
from crypteia import Multibasing, Multihashing, ToBytes, compose
from fastapi import HTTPException
from redbaby.behaviors import ReadingMixin
from redbaby.pyobjectid import PyObjectId

from ..schemas import Infostar
from ..settings import Settings

T = TypeVar("T", bound=ReadingMixin)


def read_many(infostar: Infostar, model: Type[T], **filters) -> list[T]:
    query = {k: v for k, v in filters.items() if v is not None}
    objs = model.find(
        filter=query,
        alias=Settings.get().TAUTH_REDBABY_ALIAS,
        validate=True,
        lazy=False,
    )
    return objs


def read_one(infostar: Infostar, model: Type[T], identifier: PyObjectId | str) -> T:
    if isinstance(identifier, str):
        identifier = PyObjectId(identifier)
    filters = {"_id": identifier}
    item = model.collection(alias=Settings.get().TAUTH_REDBABY_ALIAS).find_one(filters)
    if not item:
        d = {
            "error": "DocumentNotFound",
            "msg": f"Document with filters={filters} not found.",
        }
        raise HTTPException(status_code=404, detail=d)
    item = model.model_validate(item)
    return item


def get_cacheia_key(infostar: Infostar, model: Type[T], **filters) -> str:
    key = f"{model.__name__}"
    # key += f"__{infostar.model_dump(exclude={"request_id"})}"
    key += f"__{filters}"
    get_hash = compose(ToBytes(), Multihashing("sha3-224"), Multibasing("base58btc"))
    digest = get_hash(key)
    return digest


# @cache(key_builder=get_cacheia_key)
def read_one_filters(infostar: Infostar, model: Type[T], **filters) -> T:
    f = {k: v for k, v in filters.items() if v is not None}
    items = model.find(
        f,
        alias=Settings.get().TAUTH_REDBABY_ALIAS,
        validate=True,
        lazy=False,
    )
    if not items:
        d = {
            "error": "DocumentNotFound",
            "msg": f"Document with filters={filters} not found.",
        }
        raise HTTPException(status_code=404, detail=d)
    if len(items) > 1:
        d = {
            "error": "DocumentNotUnique",
            "msg": f"Document with filters={filters} not unique.",
        }
        raise HTTPException(status_code=409, detail=d)

    return items[0]



from functools import lru_cache

from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    # API
    HOST: str = "0.0.0.0"
    PORT: int = 5000
    RELOAD: bool = False
    WORKERS: int = 1

    # Database
    TAUTH_MONGODB_DBNAME: str = "tauth"
    TAUTH_MONGODB_URI: str = "mongodb://localhost:27017/"
    TAUTH_REDBABY_ALIAS: str = "tauth"

    # Security
    TAUTH_ROOT_API_KEY: str = "MELT_/--default--1"

    model_config = SettingsConfigDict(extra="ignore", env_file=".env")

    @classmethod
    @lru_cache(maxsize=1)
    def get(cls):
        return cls()



from typing import Literal

from pydantic import BaseModel


class AuthorizationPolicyIn(BaseModel):
    description: str
    name: str
    policy: str
    type: Literal["opa"]


class AuthorizationDataIn(BaseModel):
    context: dict
    policy_name: str
    resource: str


from fastapi.openapi.models import Example

POLICY_EXAMPLES = {
    "opa_melt_key": Example(
        summary="MELT API Key Authorization Policy",
        description="This policy is used to distinguish/authorize MELT API key users.",
        value=dict(
            type="opa",
            name="melt-key",
            description="MELT API Key privilege levels.",
            policy="""
package tauth.melt_key

import rego.v1

default is_valid_user = false
default is_valid_admin = false
default is_valid_superuser = false

is_valid_user := true if {
    input.infostar.authprovider_type == "melt-key"
}

is_valid_admin := true if {
    is_valid_user
    input.infostar.apikey_name == "default"
}

is_valid_superuser := true if {
    is_valid_admin
    input.infostar.authprovider_org == "/"
}
""",
        ),
    )
}



from fastapi import HTTPException
from fastapi import status as s
from http_error_schemas.schemas import RequestValidationError

from .token import sanitize_client_name


def validate_creation_access_level(client_name: str, creator_client_name: str):
    """
    Validate if `client_name` is a direct child of `creator_client_name`.

    >>> validate_creation_access_level("/teia/athena/", "/teia")
    None
    >>> validate_creation_access_level("/teia", "/")
    None
    >>> validate_creation_access_level("/teia", "/osf")
    fastapi.exceptions.HTTPException: 403 Forbidden
    >>> validate_creation_access_level("/teia/athena", "/")
    fastapi.exceptions.HTTPException: 403 Forbidden
    """
    clean_client_name = sanitize_client_name(client_name)
    parent_name = clean_client_name.rsplit("/", 1)[0]
    if parent_name == "":
        parent_name = "/"
    if parent_name != creator_client_name:
        m = f"Client {creator_client_name!r} cannot create client {client_name!r}."
        raise HTTPException(status_code=s.HTTP_403_FORBIDDEN, detail=m)


def validate_scope_access_level(client_name: str, creator_client_name: str):
    clean_client_name = sanitize_client_name(client_name, loc=["path", "client_name"])
    # Allow / to operate in /client and /client/subclient, but not the opposite
    if clean_client_name.find(creator_client_name) != 0:
        details = RequestValidationError(
            loc=["path", "client_name"],
            msg=f"Cannot operate on token {clean_client_name!r} which lies outside or above your access level ({creator_client_name!r}).",
            type="InvalidClientName",
        )
        raise HTTPException(status_code=s.HTTP_403_FORBIDDEN, detail=details)



from pymongo import IndexModel
from redbaby.behaviors.hashids import HashIdMixin
from redbaby.behaviors.reading import ReadingMixin
from redbaby.document import Document

from ...utils.teia_behaviors import Authoring


class TokenDAO(Document, Authoring, ReadingMixin, HashIdMixin):
    client_name: str
    name: str
    value: str

    @classmethod
    def collection_name(cls) -> str:
        return "melt-keys"

    @classmethod
    def indexes(cls) -> list[IndexModel]:
        idxs = [IndexModel([("client_name", 1), ("name", 1)], unique=True)]
        return idxs

    def hashable_fields(self) -> list[str]:
        return [self.client_name, self.name]



from pathlib import Path

from fastapi import APIRouter, Body, Depends, Query, Request
from fastapi import status as s
from loguru import logger

from ...authz import privileges
from ...schemas import Infostar
from ...schemas.gen_fields import GeneratedFields
from ...utils import reading
from ..policies.models import AuthorizationPolicyDAO
from ..policies.schemas import AuthorizationPolicyIn
from . import controllers as authz_controller
from .schemas import POLICY_EXAMPLES

service_name = Path(__file__).parents[1].name
router = APIRouter(prefix=f"/{service_name}", tags=[service_name + " 🔐"])


# TODO: read_one, update_one, delete_one


@router.put("/policies", status_code=s.HTTP_201_CREATED)
@router.put("/policies/", status_code=s.HTTP_201_CREATED, include_in_schema=False)
async def upsert_one(
    request: Request,
    body: AuthorizationPolicyIn = Body(openapi_examples=POLICY_EXAMPLES),
    infostar: Infostar = Depends(privileges.is_valid_admin),
) -> GeneratedFields:
    result = authz_controller.upsert_one(body, infostar)
    return result


@router.get("/policies/{id}", status_code=s.HTTP_200_OK)
@router.get("/policies/{id}/", status_code=s.HTTP_200_OK, include_in_schema=False)
async def read_one(
    request: Request,
    id: str,
    infostar: Infostar = Depends(privileges.is_valid_user),
) -> AuthorizationPolicyDAO:
    result = authz_controller.read_one(id)
    return result


@router.get("/policies", status_code=s.HTTP_200_OK)
@router.get("/policies/", status_code=s.HTTP_200_OK, include_in_schema=False)
async def read_many(
    request: Request,
    infostar: Infostar = Depends(privileges.is_valid_user),
) -> list[AuthorizationPolicyDAO]:
    filters: dict = {k: v for k, v in request.query_params.items() if v is not None}
    result = authz_controller.read_many(filters, infostar)
    return result


@router.delete("/policies/{id}", status_code=s.HTTP_204_NO_CONTENT)
@router.delete(
    "/policies/{id}/", status_code=s.HTTP_204_NO_CONTENT, include_in_schema=False
)
async def delete_one(
    request: Request,
    id: str,
    infostar: Infostar = Depends(privileges.is_valid_admin),
) -> None:
    result = authz_controller.delete_one(id)
    return result



from typing import Literal, Optional

from pydantic import Field
from pymongo import IndexModel
from redbaby.behaviors.objectids import ObjectIdMixin
from redbaby.behaviors.reading import ReadingMixin
from redbaby.document import Document

from ..entities.models import OrganizationRef, ServiceRef
from ..schemas.attribute import Attribute
from ..utils.teia_behaviors import Authoring


class AuthProviderDAO(Document, Authoring, ObjectIdMixin, ReadingMixin):
    external_ids: list[Attribute] = Field(
        default_factory=list
    )  # dynamic provider selection: issuer, audience
    extra: list[Attribute] = Field(
        default_factory=list
    )  # url, client_id, client_secret
    organization_ref: OrganizationRef
    service_ref: Optional[ServiceRef]  # TODO: AZP claim
    type: Literal["auth0", "melt-key", "tauth-key"]

    @classmethod
    def collection_name(cls) -> str:
        return "authproviders"

    @classmethod
    def indexes(cls) -> list[IndexModel]:
        idxs = [
            IndexModel(
                [("organization_ref.handle", 1), ("type", 1), ("service_ref", 1)],
                unique=True,
            ),
            IndexModel([("external_ids.name", 1), ("external_ids.value", 1)]),
        ]
        return idxs

    def get_external_id(self, name: str) -> Optional[str]:
        if not hasattr(self, "_ext_ids"):
            self._ext_ids = {item.name: item.value for item in self.external_ids}
        return self._ext_ids.get(name)
